"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.saving.object_registration import (
    CustomObjectScope as CustomObjectScope,
)
from synalinks.src.saving.object_registration import (
    CustomObjectScope as custom_object_scope,
)
from synalinks.src.saving.object_registration import (
    get_custom_objects as get_custom_objects,
)
from synalinks.src.saving.object_registration import (
    get_registered_name as get_registered_name,
)
from synalinks.src.saving.object_registration import (
    get_registered_object as get_registered_object,
)
from synalinks.src.saving.object_registration import (
    register_synalinks_serializable as register_synalinks_serializable,
)
from synalinks.src.saving.serialization_lib import (
    deserialize_synalinks_object as deserialize_synalinks_object,
)
from synalinks.src.saving.serialization_lib import (
    serialize_synalinks_object as serialize_synalinks_object,
)
