"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.initializers import deserialize as deserialize
from synalinks.src.initializers import get as get
from synalinks.src.initializers import serialize as serialize
from synalinks.src.initializers.empty_initializer import Empty as Empty
from synalinks.src.initializers.empty_initializer import Empty as empty
from synalinks.src.initializers.initializer import Initializer as Initializer
