"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.datasets.hotpotqa import get_input_data_model as get_input_data_model
from synalinks.src.datasets.hotpotqa import (
    get_knowledge_data_model as get_knowledge_data_model,
)
from synalinks.src.datasets.hotpotqa import get_output_data_model as get_output_data_model
from synalinks.src.datasets.hotpotqa import load_data as load_data
from synalinks.src.datasets.hotpotqa import load_knowledge as load_knowledge
